(* $Id: godi_file.ml 830 2008-06-15 11:37:34Z gerd $  
 * ----------------------------------------------------------------------
 * Copyright 2008 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
*)

open Godi_file_helper
open Printf

type split_filename =
  [ `Relative of string list
  | `POSIX_absolute of string list
  | `WIN32_drive of string * string list
  | `WIN32_remote of string * string list
  | `Error
  ]

type nameclass =
    [ `Relative | `POSIX_absolute | `WIN32_drive | `WIN32_remote | `Error ]


type filename =
    { inputname : string;
      slashname : string;
      splitname : split_filename;
      win32name_resolve : unit -> string;
      win32name : unit -> string;
    }

let mt_lz =
  lazy(get_mount_table())


let chdir_count = ref 0
  (* This is incremented by every chdir *)


let win32name splitname resolve_final_symlink =
  let cache = ref None in
  (fun () ->
     let resolve_flag =
       match !cache with
	 | None -> true
	 | Some (c, n) -> c = !chdir_count in
     if resolve_flag then (
       let (n,_) = 
	 translate_to_win32
	   (Lazy.force mt_lz) splitname resolve_final_symlink in
       cache := Some(!chdir_count, n)
     );
     match !cache with
       | None -> assert false
       | Some(_, n) -> n
  )
  


let filename s =
  let s' = slashify s in
  let splitname = get_splitname s' in
  { inputname = s;
    slashname = s';
    splitname = splitname;
    win32name_resolve = win32name splitname true;
    win32name = win32name splitname false;
  }

let inputname fn = fn.inputname
let slashname fn = fn.slashname

let nativename ?(resolve_final_symlink=true) fn =
  ( if resolve_final_symlink then fn.win32name_resolve else fn.win32name ) ()


let nameclass fn =
  match fn.splitname with
    | `Error -> `Error
    | `POSIX_absolute _ -> `POSIX_absolute
    | `Relative _ -> `Relative
    | `WIN32_drive(_,_) -> `WIN32_drive
    | `WIN32_remote(_,_) -> `WIN32_remote

let split_filename fn = fn.splitname

let combine_filename_str =
  function
    | `POSIX_absolute path ->
	"/" ^ String.concat "/" path
    | `Relative path ->
        String.concat "/" path
    | `WIN32_drive(letter,path) ->
	letter ^ ":/" ^ String.concat "/" path
    | `WIN32_remote(host,path) ->
	"//" ^ host ^ "/" ^ String.concat "/" path
    | `Error ->
        invalid_arg "Godi_file.combine_filename: unsupported name"

let concat_sn f1 f2 =
  match f2.splitname with
    | `POSIX_absolute _ | `WIN32_drive(_,_) | `WIN32_remote(_,_) ->
	f2.splitname
    | `Relative p2 ->
	( match f1.splitname with
	    | `POSIX_absolute p1 ->
		`POSIX_absolute(p1 @ p2)
	    | `WIN32_drive(letter,p1) ->
		`WIN32_drive(letter, p1 @ p2)
	    | `WIN32_remote(host,p1) ->
		`WIN32_remote(host, p1 @ p2)
	    | `Relative p1 ->
		`Relative (p1 @ p2)
	    | `Error ->
		invalid_arg "Godi_file.concat: unsupported name"
	)
    | `Error ->
        invalid_arg "Godi_file.concat: unsupported name"


let combine_filename sn =
  let slashname = combine_filename_str sn in
  { inputname = slashname;
    slashname = slashname;
    splitname = sn;
    win32name_resolve = win32name sn true;
    win32name = win32name sn false;
  }


let posixify fn =
  match fn.splitname with
    | `WIN32_drive _
    | `WIN32_remote _ ->
	let r = fn.win32name_resolve() in
	let rs = get_splitname r in
	let p_opt = translate_to_posix (Lazy.force mt_lz) rs in
	( match p_opt with
	    | None -> fn  (* No translation possible *)
	    | Some (p,_) -> combine_filename (`POSIX_absolute p)
	)
    | _ ->
	fn


let concat f1 f2 =
  combine_filename (concat_sn f1 f2)

let addname_sn f1 p2_str =
  let p2 =
    List.filter (fun n -> n <> "") (Str.split_delim pathsep_re p2_str) in
  match f1.splitname with
    | `POSIX_absolute p1 -> `POSIX_absolute(p1 @ p2)
    | `WIN32_drive(letter, p1) -> `WIN32_drive(letter, p1 @ p2)
    | `WIN32_remote(host, p1) -> `WIN32_remote(host, p1 @ p2)
    | `Relative p1 -> `Relative(p1 @ p2)
    | `Error -> `Error


let addname f1 p2_str =
  combine_filename(addname_sn f1 p2_str)


let rec list_addsuffix l suff =
  match l with
    | [] -> invalid_arg "Godi_file.addsuffix"
    | [x] -> [x ^ suff]
    | x :: l' -> x :: list_addsuffix l' suff


let addsuffix_sn f1 suff =
  match f1.splitname with
    | `POSIX_absolute p1 -> `POSIX_absolute(list_addsuffix p1 suff)
    | `WIN32_drive(letter, p1) -> `WIN32_drive(letter, list_addsuffix p1 suff)
    | `WIN32_remote(host, p1) -> `WIN32_remote(host, list_addsuffix p1 suff)
    | `Relative p1 -> `Relative(list_addsuffix p1 suff)
    | `Error -> `Error


let addsuffix f1 suff =
  combine_filename (addsuffix_sn f1 suff)


let rec list_basename l =
  match l with
    | [] -> "/"
    | [x] -> x
    | _ :: l' -> list_basename l'


let basename f1 =
  match f1.splitname with
    | `POSIX_absolute p1 -> list_basename p1
    | `WIN32_drive(letter, p1) -> list_basename p1
    | `WIN32_remote(host, p1) -> list_basename p1
    | `Relative p1 -> list_basename p1
    | `Error -> invalid_arg "Godi_file.basename"


let rec list_dirname l =
  match l with
    | [] -> raise Not_found
    | [_] -> []
    | x :: l' -> x :: list_dirname l'


let dirname f1 =
  try
    let sn =
      match f1.splitname with
	| `POSIX_absolute p1 -> 
	    `POSIX_absolute(list_dirname p1)
	| `WIN32_drive(letter, p1) -> 
	    `WIN32_drive(letter, (list_dirname p1))
	| `WIN32_remote(host, p1) -> 
	    `WIN32_remote(host, (list_dirname p1))
	| `Relative p1 ->
	    let p1' = list_dirname p1 in
	    `Relative(if p1' = [] then [ "." ] else p1')
	| `Error -> invalid_arg "Godi_file.dirname"
    in
    Some(combine_filename sn)
  with
    | Not_found -> None
	


let q_re = Str.regexp "'"

let uquote s = 
  "'" ^ Str.global_replace q_re "'\\''" s ^ "'"


exception Not_a_prefix

let strip_prefix ~prefix file =
  let rec strip p f =
    match (p, f) with
      | (cp :: p'), (cf :: f') when cp = cf ->
          strip p' f'
      | [], _ ->
          f
      | _ ->
          raise Not_a_prefix
  in

  match (prefix,file) with
    | `POSIX_absolute p, `POSIX_absolute f ->
        strip p f
    | `Relative p, `Relative f ->
        strip p f
    | `WIN32_drive(pd,p), `WIN32_drive(fd,f) ->
	if String.lowercase pd = String.lowercase fd then 
	  strip p f
	else raise Not_a_prefix
    | `WIN32_remote(pd,p), `WIN32_remote(fd,f) ->
	(* CHECK: case-insensitive comparison? *)
	if pd = fd then strip p f else raise Not_a_prefix
    | _ ->
        raise Not_a_prefix


type mode = [ `Text | `Binary ]

let open_in =
  function
    | `Text -> (fun n -> Pervasives.open_in (nativename n))
    | `Binary -> (fun n -> Pervasives.open_in_bin (nativename n))

let open_in_gen flags mode n = 
  Pervasives.open_in_gen flags mode (nativename n)

let open_out =
  function
    | `Text -> (fun n -> Pervasives.open_out (nativename n))
    | `Binary -> (fun n -> Pervasives.open_out_bin (nativename n))

let open_out_gen flags mode n = 
  Pervasives.open_out_gen flags mode (nativename n)


let catch_unix_error name f arg =
  try f arg
  with
    | Unix.Unix_error(err,_,errarg) ->
        let errarg' = if errarg <> "" then ": " ^ errarg else errarg in
        raise(Sys_error(Unix.error_message err ^ errarg' ^ " (" ^ name ^ ")" ))
;;


let getcwd () =
  try 
    let cwd = slashify (Unix.getcwd()) in
    let cwd_splitname = get_splitname cwd in
    ( match translate_to_posix (Lazy.force mt_lz) cwd_splitname with
	| None ->
	    Some
	      { inputname = cwd;
		slashname = cwd;
		splitname = cwd_splitname;
		win32name_resolve = (fun () -> cwd);
		win32name = (fun () -> cwd);
	      }
	| Some(p,_) ->
	    let splitname = `POSIX_absolute p in
	    let s = combine_filename_str splitname in
	    Some
	      { inputname = s;
		slashname = s;
		splitname = splitname;
		win32name_resolve = (fun () -> cwd);
		win32name = (fun () -> cwd);
	      }
    )
  with
    | e ->
        catch_unix_error "getcwd" (fun () -> raise e) ()


let chdir =
  catch_unix_error
    "chdir"
    (fun n -> 
       incr chdir_count;
       Unix.chdir (nativename n)
    )


type file_id =
    { file_device : int32;
      file_inode : int64;
    }

let file_id ?(symlink=false) =
  catch_unix_error
    "file_id"
    (fun file ->
       let n =
	 nativename ~resolve_final_symlink:(not symlink) file in
       let (device, inode) = Godi_win32.get_identity_information n in
       { file_inode = inode;
         file_device = device;
       }
    )
;;


let same_file f1 f2 =
  (* Returns [true] if both files exist, and both files are the same *)
  try
    let id1 = file_id f1 in
    let id2 = file_id f2 in
    id1 = id2
  with
    | Sys_error _ -> false
;;


type file_fingerprint =
    { file_mtime : float;
      file_ctime : float;
      file_size : int;
    }
;;


let file_fingerprint =
  catch_unix_error
    "file_fingerprint"
    (fun f ->
       let s = Unix.stat (nativename f) in
       { file_mtime = s.Unix.st_mtime;
         file_ctime = s.Unix.st_ctime;
         file_size = s.Unix.st_size;
       }
    )
;;


let subst_symlinks d l =
  (* If the dir listing l contains symlinks, we have to strip the ".lnk" 
     suffix off
   *)
  List.map
    (fun n ->
       if Str.string_match lnk_re n 0 then (
	 match
	   read_shortcut_symlink
	     (nativename ~resolve_final_symlink:false (addname d n))
	 with
	   | Some _ ->
	       Filename.chop_extension n (* it is a symlink *)
	   | None ->
	       n   (* no symlink despite the .lnk suffix *)
       )
       else n
    )
    l


let read_dir ?(include_dots=false) =
  (* Sys.readdir does not return ".", and "..", so we do not use it *)
  catch_unix_error
    "read_dir"
    (fun d ->
       let dh = Unix.opendir (nativename d) in
       let f = ref [] in
       try
         while true do
           let n = Unix.readdir dh in
           if (n <> "." && n <> "..") || include_dots then
             f := n :: !f;
         done;
         assert false
       with
           End_of_file ->
             Unix.closedir dh;
             subst_symlinks d (List.rev !f)
         | error ->
             Unix.closedir dh;   (* Cannot read entry *)
             raise error
    )
;;


let stat ?(symlink=false) =
  catch_unix_error
    "stat"
    (fun file ->
       if symlink then (
	 let n = nativename ~resolve_final_symlink:false file in
	 ( match read_shortcut_symlink n with
	     | Some target ->
		 let st = Unix.LargeFile.stat n in
		 { st with
		     Unix.LargeFile.st_kind = Unix.S_LNK;
		     st_perm = 0o777;
		     st_size = Int64.of_int(String.length target)
		 }
	     | None ->
		 Unix.LargeFile.stat n
	 )
       )
       else (
	 let n = nativename file in
	 Unix.LargeFile.stat n
       )
    )


let file_type ?(is_regular = false) 
              ?(is_dir = false) 
              ?(is_existing = false)
              ?(is_symlink = false) 
              ?(is_not_symlink = false) 
              fn =
  let lkind =
    try
      Some ((stat ~symlink:true fn).Unix.LargeFile.st_kind)
    with
        Sys_error _ -> None in
  let kind =
    try
      Some ((stat fn).Unix.LargeFile.st_kind)
    with
        Sys_error _ -> None in
  (* Negative logic: *)
  (* Note that there are abnormal cases when "file" and "file.lnk" both exist.
     We do nothing about that.
   *)
  match lkind, kind with
      Some Unix.S_REG, _ -> not is_dir && not is_symlink
    | Some Unix.S_DIR, _ -> not is_regular && not is_symlink
    | Some Unix.S_LNK, Some Unix.S_REG -> not is_dir && not is_not_symlink
    | Some Unix.S_LNK, Some Unix.S_DIR -> not is_regular && not is_not_symlink
    | Some Unix.S_LNK, Some _ -> 
        not is_regular && not is_dir && not is_not_symlink
    | Some Unix.S_LNK, None ->
        not is_regular && not is_dir && not is_existing && not is_not_symlink
    | Some _, _ ->
        not is_regular && not is_dir && not is_symlink
    | None, _ ->
        not is_regular && not is_dir && not is_existing && not is_symlink
;;


let file_exists ?(symlink=false) fn =
  file_type ~is_existing:true fn || (symlink && file_type ~is_symlink:true fn)
;;


let file_is_dir fn =
  file_type ~is_dir:true fn ;;


            
let file_mtime ?(symlink=false) fn =
  (stat ~symlink fn).Unix.LargeFile.st_mtime


let set_mtime fn =
  (* Note: this function never targets symlinks *)
  catch_unix_error
    "set_mtime"
    (fun t -> 
       Unix.utimes (nativename fn) t t)


let set_mode fn =
  catch_unix_error
    "set_mode"
    (Unix.chmod (nativename fn))


            
let file_size fn =
  (stat fn).Unix.LargeFile.st_size
;;


let rename oldname =
  catch_unix_error
    "rename"
    (fun newname ->
       let old_st =
	 stat ~symlink:true oldname in
       let old_lnk =
	 old_st.Unix.LargeFile.st_kind = Unix.S_LNK in
       let new_lnk =
	 try 
	   Some 
	     ((stat ~symlink:true newname).Unix.LargeFile.st_kind = Unix.S_LNK )
	 with Sys_error _ -> None in
       let n_oldname = nativename ~resolve_final_symlink:false oldname in
       let n_newname = nativename ~resolve_final_symlink:false newname in
       (* The ~symlink:true param prevents that we resolve the last component
          of the names as symlink. Nevertheless, an ".lnk" suffix may have
          been added! Also note that symlinks are write-protected!
	*)
       if old_lnk then (
	 Unix.chmod n_oldname 0o666;
	 match new_lnk with
	   | None ->
	       let n_newlnk = n_newname ^ ".lnk" in
	       Unix.rename n_oldname n_newlnk;
	       Unix.chmod n_newlnk 0o444
	   | Some false ->
	       let n_newlnk = n_newname ^ ".lnk" in
	       Unix.unlink n_newname;
	       Unix.rename n_oldname n_newlnk;
	       Unix.chmod n_newlnk 0o444
	   | Some true ->
	       Unix.rename n_oldname n_newname;
	       Unix.chmod n_newname 0o444
       )
       else (
	 if new_lnk = Some true then (
	   Unix.chmod n_newname 0o666;
	   Unix.unlink n_newname;
	 );
	 (* Get the newname without ".lnk" suffix: *)
	 let n_newname = nativename ~resolve_final_symlink:false newname in
	 (* Unix allows it to rename write-protected files. So we allow it 
            too
	  *)
	 Unix.chmod n_oldname 0o666;
	 Unix.rename n_oldname n_newname;
	 Unix.chmod n_newname old_st.Unix.LargeFile.st_perm
       )
    )
;;



let read_dir_hierarchy ?(include_dirs=false) root =
  let rec collect rel_path =
    let path = addname root rel_path in
    let files = read_dir path in
    List.flatten
      (List.map
         (fun file ->
            let p = addname path file in
            let r = if rel_path <> "" then rel_path ^ "/" ^ file else file in
            if file_type ~is_dir:true ~is_not_symlink:true p then (
              let sub_dir = collect r in
              if include_dirs then r :: sub_dir else sub_dir
            )
            else [ r ]
         )
         files)
  in
  collect ""
;;


let read_file_bin filename =
  (* Return the contents of [filename] *)
  let f = open_in `Binary filename in
  try
    let n = in_channel_length f in
    let s = String.create n in
    really_input f s 0 n;
    close_in f;
    s
  with
      error -> close_in f; raise error
;;


let crlf_re = Str.regexp "\r\n";;
let lf_re = Str.regexp "\n";;

let read_file ~mode filename =
  let s = read_file_bin filename in
  match mode with
    | `Binary -> s
    | `Text -> Str.global_replace crlf_re "\n" s
;;


let read_file_lines filename =
  let rec fixup_eof lines =
    match lines with
      | [ "" ] -> []   (* Last line ends with LF *)
      | [] -> []
      | line :: lines -> line :: fixup_eof lines
  in
  fixup_eof (Str.split_delim lf_re (read_file ~mode:`Text filename))
;;


let write_file ~mode filename contents =
  let f = open_out_gen [ Open_wronly; Open_creat; Open_trunc;
                         ( match mode with
                             | `Text -> Open_text
                             | `Binary -> Open_binary
                         )
                       ] 0o666 filename in
  try
    output_string f contents;
    close_out f
  with
    | error -> close_out f; raise error
;;


let write_file_lines filename lines =
  let f = 
    open_out_gen 
      [ Open_wronly; Open_creat; Open_trunc; Open_text ] 0o666 filename in
  try
    List.iter
      (fun line ->
         output_string f line;
         output_string f "\r\n"
      )
      lines;
    close_out f
  with
    | error -> close_out f; raise error
;;


let rec remove ?(recurse = false) =
  let unlink d =
    let n = nativename ~resolve_final_symlink:false d in
    Unix.chmod n 0o666;
    Unix.unlink n
  in

  catch_unix_error
    "remove"
    (fun d ->
       if recurse then (
         if file_type ~is_dir:true ~is_not_symlink:true d then (
           let l = read_dir d in
           List.iter
             (fun f ->
                let df = addname d f in
                if f <> "." && f <> ".." then (
                  remove ~recurse df
                )
             )
             l;
           Unix.rmdir (nativename d)
         )
         else
           if file_exists ~symlink:true d then
	     unlink d
       )
       else
         if file_type ~is_dir:true ~is_not_symlink:true d then
           Unix.rmdir (nativename d)
         else
	   unlink d
    )
;;


let remove_dir ?(recurse = false) fn =
  if file_type ~is_dir:true ~is_not_symlink:true fn then
    remove ~recurse fn
  else
    raise(Sys_error("Not a directory: " ^ slashname fn))
;;


exception Symlink_in_the_way

let rec create_dir ?(recurse = false) =
  catch_unix_error
    "create_dir"
    (fun d ->
       if recurse then (
         try
	   let is_symlink =
	     try (stat ~symlink:true d).Unix.LargeFile.st_kind = Unix.S_LNK 
	     with Sys_error _ -> false in
	   if is_symlink then
	     raise Symlink_in_the_way;
           Unix.mkdir (nativename d) 0o777
         with
	   | Symlink_in_the_way
           | Unix.Unix_error(Unix.EEXIST,_,_) ->
               (* Only check that it is already a directory: *)
               if not (file_type ~is_dir:true d) then
                 raise(Sys_error ("Not a directory: " ^ slashname d))
           | Unix.Unix_error(Unix.ENOENT,_,_) as e ->
               (* Try to ensure that the parent exists: *)
	       ( match dirname d with
		   | Some d' ->
		       create_dir ~recurse:true d';
		       Unix.mkdir (nativename d) 0o777
		   | None ->
		       raise e
	       )
       ) else
         Unix.mkdir (nativename d) 0o777
    )
;;


exception Not_a_symlink

let read_symlink file =
  match
    read_shortcut_symlink
      (nativename ~resolve_final_symlink:false file)
  with
    | Some target ->
	target
    | None ->
	raise Not_a_symlink


let create_symlink src dest =
  (* Currently not implemented as built-in function. Instead we call the
     Cygwin "ln" command. Not really fast, but symlinks occur only seldom
     in Godi.
   *)
  let dest_exists =
    try 
      ignore(stat ~symlink:true dest); true
    with Sys_error _ -> false in
  if dest_exists then
    raise(Sys_error ("File exists: " ^ slashname dest));

  let ln = "/bin/ln" in
  let n_ln = backslashify (nativename (filename ln)) in
  let code =
    (* Remember that Sys.command is executed by cmd.exe. We have to
       quote the args so they aren't expanded by ln. "2>nul" discards any
       error messages.
     *)
    Sys.command
      (sprintf "%s -s \"%s\" \"%s\" 2>nul"
	 n_ln
	 (slashname src)
	 (slashname dest)
      ) in
  if code <> 0 then
    raise(Sys_error(ln ^ " fails: " ^ slashname dest))
	 


let create_link src =
  catch_unix_error
    "create_link"
    (fun dest ->
       let src_lnk =
	 (stat ~symlink:true src).Unix.LargeFile.st_kind = Unix.S_LNK in
       let dest_exists =
	 try 
	   ignore(stat ~symlink:true dest); true
	 with Sys_error _ -> false in
       let n_src = nativename ~resolve_final_symlink:false src in
       let n_dest = nativename ~resolve_final_symlink:false dest in

       if dest_exists then
	 raise(Unix.Unix_error(Unix.EEXIST, "", n_dest));

       if src_lnk then
	 Unix.link n_src (n_dest ^ ".lnk")
       else
	 Unix.link n_src n_dest
    )


let tmpdir_dir = ref None;;
let tmpdir_count = ref 0 ;;

type tmpdir = filename ;;

let get_tmpdir_state() =
  ( match !tmpdir_dir with
      | None -> ""
      | Some d -> d
  ) ^ ":" ^ string_of_int !tmpdir_count

let state_re = Str.regexp "\\([^:]*\\):\\([0-9]\\)$"

let init_tmpdir_state s =
  if !tmpdir_dir <> None then
    invalid_arg "Godi_file.init_tmpdir_state: already initialized";
  if Str.string_match state_re s 0 then (
    let d = Str.matched_group 1 s in
    if d <> "" then
      tmpdir_dir := Some d;
    tmpdir_count := int_of_string(Str.matched_group 2 s);
  )
  else
    invalid_arg "Godi_file.init_tmpdir_state: parse error"


let tmpdir_path t = t ;;

let rec create_tdir prefix n =
  try
    let name = prefix ^ string_of_int n in
    Unix.mkdir name 0o777;
    n
  with
      Unix.Unix_error(Unix.EEXIST,_,_) ->
        create_tdir prefix (n+1)
;;


let create_tmp_dir() =
  let tmp_prefix = "/tmp/godi" ^ string_of_int (Unix.getpid()) ^ "-" in
  let tmp_native = nativename (filename tmp_prefix) in
  let n = catch_unix_error "create_tmp_dir" (create_tdir tmp_native) 0 in
  filename(tmp_prefix ^ string_of_int n)
;;


let allocate_tmpdir() =
  (* Ensure tmpdir_dir exists: *)
  let tmpdir_loc =
    match !tmpdir_dir with
        None ->
          let tmp_prefix = "/tmp/godi" ^ string_of_int (Unix.getpid()) ^ "-" in
	  let tmp_native = nativename (filename tmp_prefix) in
          let n = catch_unix_error "allocate_tmpdir" (create_tdir tmp_native) 0 in
	  let f = filename(tmp_prefix ^ string_of_int n) in
          tmpdir_dir := Some f;
          f
      | Some d -> d
  in
  (* Fetch next number: *)
  let n = !tmpdir_count  in
  incr tmpdir_count;
  (* Create directory: *)
  let name = addname tmpdir_loc (string_of_int n) in
  create_dir name;
  name
;;


let release_tmpdir dir =
  remove_dir ~recurse:true dir
;;


let release_all_tmpdirs() =
  match !tmpdir_dir with
      None -> ()
    | Some d ->
        remove_dir ~recurse:true d;
        tmpdir_dir := None
;;


let copy ?(prefer_link = false) ?remove:(rm = false) origname copyname =
  (* We ignore prefer_link because this is totally unexpected by Windows
     users
   *)

  catch_unix_error
    "copy"
    (fun () ->
       let b = String.create 4096 in
       let origfd = open_in `Binary origname in
       try
         if rm && file_exists copyname then
           remove copyname;
         let copyfd = open_out `Binary copyname in
         ( try
             let eof = ref false in
             while not !eof do
               let n = input origfd b 0 (String.length b) in
               if n=0 then eof := true;
               output copyfd b 0 n
             done;
	     (* No fstat in Windows, ... so use stat *)
             let p = (stat origname).Unix.LargeFile.st_perm in
               set_mode copyname (p land 0o777);
             close_in origfd;
             close_out copyfd;
           with
               error ->
                 close_out copyfd; raise error
         )
       with
           error -> 
             close_in origfd; raise error
    )
    ()


let percent_s_re = Str.regexp "%s"

let execute ?(setx=false) cmd =
  let cmd_exe =
    try Sys.getenv "COMSPEC"
    with Not_found -> "cmd.exe" in
  let tmpdir = allocate_tmpdir() in  (* Do this before getting the state! *)
  let state = get_tmpdir_state() in
  let cmd = Str.global_replace percent_s_re state cmd in
  try
    let tmpdir_path = tmpdir_path tmpdir in
    let script = addname tmpdir_path "script" in
    let f = open_out `Binary script in
    output_string f cmd;
    close_out f;
    let switches = if setx then ["-x"] else [] in
    let bin_sh = nativename(filename "/bin/ash") in
    Unix.execv
      cmd_exe 
      (Array.of_list
         (["/d"; "/c"; bin_sh] @ switches @ [slashname script]));
  with
      Unix.Unix_error(err, _, param) ->
        let s = Unix.error_message err in
        let s' = if param="" then s else s ^ ": " ^ param in
        raise(Sys_error s')

